// OSIKCS.cpp : Defines the entry point for the console application.
// Mark Spankus 2020
//
#include "osikcs.h"

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdarg.h>
//#include <io.h>
#include <vector>


FILE* logfile = stderr;
bool bVerbose = false;

bool bEven   = false;
bool bOdd    = false;
bool b7bits  = false;
bool bStop1   = false;
bool bInvert = false;
bool bSin    = false;
bool bOverwrite = false;
bool bDecode = false;
bool bBASICmode = false;
bool bAppend = false;
bool bWavDiag = false;

int  nChannel = 0; //channel to sample in multichannel input audio
int  nLeader = 2000;  //leader length in mseconds
int  nTrailer= 1000;  //trailer length in mseconds
int  nNulls  = 0;
int  nRate   = 0; //0=300 1=600
int  nPadding = 0;
int  nWordPadding = 0;
int  nFrameErrors = 0;
int  nParityErrors = 0;


#define WAVELEN 18
#define CHUNKSIZEOFFSET 40
#define FILESIZEOFFSET 4
#define SAMPLERATE 22050
#define BAUD 300


#define UNKNOWN_ERROR    -1
#define INVALID_ENCODING -2
#define INVALID_FILETYPE -3
#define BAD_FILE -4


// Riff Wave Header
typedef struct WaveHeader {    
    char chunkId[4];		//4 Chunk ID: "RIFF"
    long chunkSize;			//4 Chunk size: 4+n
    char format[4];			//4 WAVE ID: "WAVE"

    // Format Subchunk
    char subChunk1Id[4];	//4 Chunk ID: "fmt "
    long  subChunk1Size;	//4 Chunk size: 16, 18 or 40
    short audioFormat;		//2 Format code	01=PCM 03=IEEE_FLOAT 06=WAVE_ALAW 07=WAVE_MULAW 0xFFFE=determined by subformat
    short numChannels;		//2 Number of interleaved channels
    long sampleRate;		//4 Sampling rate (blocks per second)
    long byteRate;			//4 Data rate
    short blockAlign;		//2 Data block size (bytes)
    short bitsPerSample;	//2 Bits per sample
    //short extraParamSize;	 //size of extension 0 or 22 if chunk1size=18
	
	//extension	 if chunk1size=40
	//  short wValidBitsPerSample	 //Number of valid bits
	//  long dwChannelMask;  //Speaker Position mask
	//  unsigned char SubFormat[16] //GUID, including data fmt code
    
	// Data Subchunk
    char subChunk2Id[4];	 //4 "data
    long  subChunk2Size;	 //4 size
	//followed by n samples
	//and 1 byte padding if n is odd

} WaveHeader;


// WAVE file header format
struct HEADER {
	 char riff[4];						// RIFF string
	unsigned int  overall_size	;		// overall size of file in bytes
	 char wave[4];						// WAVE string
	 char fmt_chunk_marker[4];			// fmt string with trailing null char
	unsigned int  length_of_fmt;		// length of the format data
	unsigned short format_type;			// format type. 1-PCM, 3- IEEE float, 6 - 8bit A law, 7 - 8bit mu law
	unsigned short channels;			// no.of channels
	unsigned int  sample_rate;			// sampling rate (blocks per second)
	unsigned int  byterate;				// SampleRate * NumChannels * BitsPerSample/8
	unsigned short block_align;			// NumChannels * BitsPerSample/8
	unsigned short bits_per_sample;		// bits per sample, 8- 8bits, 16- 16 bits etc
	 char data_chunk_header [4];		// DATA string or FLLR string
	unsigned int  data_size;			// NumSamples * NumChannels * BitsPerSample/8 - size of the next chunk that will be read
};

//RIFF LIST
// INFO
//  INAM	  title
//  IPRD	  product
//  IART	  artist
//  ICMT	  comment
//  ICRD	  date created
//  ITRK
//  TALB
//  TDRC
//  TIT2
//  ITPE
//  COMM	 comments

unsigned char ZEROBIT_SQR_KCS[] = //  ( 1 cycle 1200 hz - square equalized )
{
	//( normal )
	0x9B, 0xC1, 0xD9, 0xE8, 0xF2, 0xF9, 0xFC, 0xFF, 0xA0,
	0x64, 0x3E, 0x26, 0x17, 0x0D, 0x06, 0x03, 0x00, 0x5F,
	//( inverted )
	0x64, 0x3E, 0x26, 0x17, 0x0D, 0x06, 0x03, 0x00, 0x5F,
	0x9B, 0xC1, 0xD9, 0xE8, 0xF2, 0xF9, 0xFC, 0xFF, 0xA0,
};

unsigned char ONEBIT_SQR_KCS[] = //( 2 cycle 2400 hz - square equalized )
{
	//normal
	0x9D, 0xDC, 0xF5, 0xFF, 0x97, 0x55, 0x2B, 0x11, 0x00,
	0x9D, 0xDC, 0xF5, 0xFF, 0x97, 0x55, 0x2B, 0x11, 0x00,
	//( inverted )
	0x62, 0x23, 0x0A, 0x00, 0x68, 0xAA, 0xD4, 0xEE, 0xFF,
	0x62, 0x23, 0x0A, 0x00, 0x68, 0xAA, 0xD4, 0xEE, 0xFF
};

unsigned char ZEROBIT_SIN_KCS[] = //  ( 1 cycle 1200 hz bit - sine )
{
	//( normal )
	0x96, 0xC0, 0xE2, 0xF8, 0xFF, 0xF8, 0xE2, 0xC0, 0x96,
	0x6A, 0x40, 0x1E, 0x08, 0x00, 0x08, 0x1E, 0x40, 0x6A,
	//( inverted )
	0x69, 0x3F, 0x1D, 0x07, 0x00, 0x07, 0x1D, 0x3F, 0x69,
	0x95, 0xBF, 0xE1, 0xF7, 0xFF, 0xF7, 0xE1, 0xBF, 0x95,
};

unsigned char ONEBIT_SIN_KCS[] =  //( 2 cycle 2400 hz - sine )
{
	//( normal )
	0xD2, 0xFE, 0xEF, 0xAC, 0x54, 0x11, 0x02, 0x2E, 0x80,
	0xD2, 0xFE, 0xEF, 0xAC, 0x54, 0x11, 0x02, 0x2E, 0x80,
	//( inverted )
	0x2D, 0x01, 0x10, 0x53, 0xAB, 0xEE, 0xFD, 0xD1, 0x7F,
	0x2D, 0x01, 0x10, 0x53, 0xAB, 0xEE, 0xFD, 0xD1, 0x7F,
};

const char WAV_HDR[] = //  ( predefined WAV header )
{
	0x52, 0x49, 0x46, 0x46, 0x00, 0x00, 0x00, 0x00,		//"RIFF	+ chunk size (filled later)
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20,	    //"WAVEfmt
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x01, 0x00,		//chunk1size(16), PCM, mono
	0x22, 0x56, 0x00, 0x00, 0x22, 0x56, 0x00, 0x00,		//22050 blocks/per second, 22050 bytes/second (std speed - a little fast for our waveform)
//	0x60, 0x54, 0x00, 0x00, 0x60, 0x54, 0x00, 0x00,		//21600 blocks/per second, 21600 bytes/second (desired speed-exactly matches our waveform)
	0x01, 0x00, 0x08, 0x00, 0x64, 0x61, 0x74, 0x61,		// blocksize, bitspersample,  "data
	0x00, 0x00, 0x00, 0x00							    // data chunk size
};

Emitter::Emitter(FILE* target)
{
	setFrame();
	setTiming();
	m_nframeLen = 11;
	m_nBitsIn = 0;
	m_nCurPos = 0;
	m_nLastPos = 0;
	m_nStart = 0;
	nJitter = 2;
	bStart = false;
	m_target = target;
	m_nWrote = 0;

	bFrameError = false;
	m_nCalcLeader = 0;
	FEcount = 0;
	PEcount = 0;
	m_Avg1200 = -1;
	m_Avg2400 = -1;
	m_freq2400 = 0.0;
	int i;
	for (i=0;i<12;i++)
		frame[i]=1;
}

Emitter::~Emitter()
{
}

bool Emitter::setTiming(int nFreq /*= 22050*/, int nBps /*= 300*/)
{
	if (nFreq < 16000 || nFreq > 65536)
		return 0;
	
	m_nFreq = nFreq;
	m_nBps = nBps;

	//these values are updated by sampling leader "2400Hz" tone, but we need ballpark initial values

	m_nFrameSize = (m_nFreq / nBps / 2 ) +4;
	nJitter = 2 +  (m_nFreq/22050);
	m_Avg1200 = m_nFreq/2400+1;	   //guestimate # of samples for 1/2 wave form
	m_Avg2400 = m_nFreq/4800 +1;   //guestimate # of samples for 1/2 wave form

	return 1;

}
void Emitter::setFrame(int nBits/* = 8*/, int nStop/* = 2*/, eParity parity/* = eNone*/)
{
	// 1 startbit (0), 7-8 data bits (d), 0-1 parity (p), 1-2 stop bits (1)
	m_nBits = nBits;
	m_nStop = nStop;
	m_parity = parity;

	//frame
	m_nframeLen = 1 + nBits + nStop + (m_parity ==  eNone ? 0: 1);

	m_nStopBit1Pos = 1 + nBits + m_parity;
	m_nStopBit2Pos = m_nStopBit1Pos;
	if (nStop > 1)
		m_nStopBit2Pos++;
	m_nParityPos = -1;
	if (m_parity)
		m_nParityPos = m_nStopBit1Pos -1;
}

int Emitter::QSum()
{
	int nSum = 0;
	for (vQitr = vQueue.begin();  vQitr != vQueue.end(); vQitr++)
	{
		nSum += *vQitr;
	}

	return nSum;
	
}

//
//add bit to stream, when serial word is complete, emit byte
//
bool Emitter::emit(bool bBit)
{
	int i;
	//shift bits over, yeah coulda used a single int instead
	for (i = 1; i < m_nframeLen; i++)
	{
		frame[i-1] = frame[i];
	}
    if (bWavDiag)
	{
		_ftprintf( logfile, _T("%c "), bBit ? _T('1') : _T('0'));
    }
	//set lowest bit
	frame[m_nframeLen-1] = bBit ? 1 : 0;
	m_nBitsIn++;
	bool bParityError= false;
	
	if (frame[0] == 0 && m_nBitsIn >= m_nframeLen )
	{
		//start bit present
		if (frame[m_nStopBit1Pos] == 1 && frame[m_nStopBit2Pos] == 1)
		{
			//Stop bits present
			unsigned char c = 0; int bits1= 0;
			for (i = m_nBits; i>0; i--)
			{
				c = (c << 1 ) | (frame[i] ? 1 : 0);
				if (frame[i]) bits1++;
			}
			if (m_nParityPos >=0)
				if (frame[m_nParityPos])
					bits1++;


			switch (m_parity)
			{
				
				case eEven:
					if (bits1 & 1)
					{
						//PE
						bParityError = true;
					}
					break;
				case eOdd:
					if (!(bits1 & 1))
					{
						//PE
						bParityError = true;
					}
					break;
				case eMark:
					if (!frame[m_nParityPos])
					{
						//PE
						bParityError = true;
					}
					break;
				case eSpace:
					if (frame[m_nParityPos])
					{
						//PE
						bParityError = true;
					}
					break;
				case eNone:
					break;
				
			}
			if (bParityError)
				PEcount++;

			putc(c, m_target);
			m_nWrote++;
			if (bWavDiag)
			{

				unsigned int u = (unsigned int) m_nLastPos;
				//print byte offset and decoded character
				_ftprintf(logfile, _T("[%08u=%02X]%c%s%s\r\n"), u, c, isprint(c) ? c : _T(' '), bParityError ? _T("[PE]") :_T(""), bFrameError ? _T("[FE]") : _T("") );
			}

			for (i=0;i<12;i++)
				frame[i]=1;	  //reset frame data

			
			m_nBitsIn = 0;
			bFrameError = false;

		}
		else
		{
			//framing error
			if (!bFrameError)
			{
				bFrameError = true;
				FEcount++;
			}
		}
	}

	return 0;
	
}

//add length count of wavform to queue
int Emitter::add(int nCycles, size_t nPos)
{

	if (!bStart && nCycles > m_nFrameSize)
		return 0;

	m_nLastPos = m_nCurPos;
	m_nCurPos = nPos;
	if (!bStart)
	{
		bStart = true;
		nCount = 0;
		
	}



	if (nCycles > m_nFrameSize) //error
	{
		return 0;
	}
	else
	{
		
		if (!m_nStart && !m_nCalcLeader)
		{
			m_nStart = nPos;
			nCount = 0;
			
		}

		nCount++;
		int size = vQueue.size();
		if (size > 0)
		{
			int nSum = QSum();
			int nAvg = nSum/size;


			if (abs(nCycles - nAvg) > nJitter )	  //at higher sample rates allow for larger jitter
			{
				//transition?
				bool bAvg1200 =  (abs(nAvg - m_Avg1200) < abs( nAvg - m_Avg2400));
				bool bCycle1200 = (abs(nCycles - m_Avg1200) < abs( nCycles - m_Avg2400));
				
			    //if the cycle time is outside of the average of the queue, then we act on it
				//otherwise its just an outlier on a small queue average
				if ( bAvg1200 ^ bCycle1200 )
				{

					//if Q size is too small, ignore transition
					if (size > (m_nBps < 500 ? 2 : 1) )
					{
						if (abs(nAvg - m_Avg1200) <= 2)
							emit(0);
						else if (abs(nAvg - m_Avg2400) <= 2)
							emit(1);
						
						else 
						if (nCycles > nAvg)
						{
							// 1->0
							emit(1);
							
						}
						else
						{
							//0->1
							emit(0);
						}
					}

					m_nStart = 0;
					vQueue.clear();
				}
			}
			else if (nSum + nCycles > m_nFrameSize )
			{

				//sequential
				if (abs(nAvg-m_Avg2400) < abs(nAvg-m_Avg1200) )
				//if (size > (m_nBps < 500 ? 5 : 3) )
				{
					//1 bit = 2xfreq of 0 bit
					//qsize is larger for 1 bit
					emit(1);
					m_Avg2400 =	nAvg;
					if (m_nStart )
					{
						size_t samples = (nPos - m_nStart);

						if (m_nCalcLeader < 2 && samples)
							m_freq2400 = (double)nCount * m_nFreq / samples;

 						
						if (nCount >  1000)	//after ~1000 cycles, frequency count is good enough to use
						{
							m_nCalcLeader = 1;
						
							int calc_frame_size =  m_nFreq * 4.5 / m_freq2400;	//4.5 2400hz cycles is max bit size
							m_nFrameSize	= (m_nBps < 500 ? calc_frame_size : (calc_frame_size/2)-1);
							m_nBps = m_freq2400 * 300 / (m_nBps < 500 ?  2400 : 1200);
						    m_Avg1200 =	2*m_Avg2400;	
						
						}
					}


				}
				else
				{

					if (m_nCalcLeader == 1)
					{
						m_nCalcLeader = 2; //stop further calculations

					}
					//0 bit
					emit(0);
				
					nCount = 0;
					m_nStart = 0;

				}

				vQueue.clear();

			}
		}


		vQueue.push_back( nCycles );

	}
	return 1;

}

//0 bit = 4 cycles of 1200Hz @ 300 baud,  2 @ 600 baud
//1 bit = 8 cycles of 2400 hz @ 300 baud, 4 @ 600 baud

size_t writeBit( bool bBit, FILE* out)
{
	size_t nReturn = 0, nVal, i, nSize = 4;
	const unsigned char * pdata;

	if (bBit)
	{
		pdata = bSin ? ONEBIT_SIN_KCS : ONEBIT_SQR_KCS;
	}
	else
	{
		pdata = bSin ? ZEROBIT_SIN_KCS : ZEROBIT_SQR_KCS;

	}
	if (nRate  )
		nSize = 2;		 // @ 600baud each cycle is 1/2 length

	if (bInvert)
		pdata += WAVELEN;

	for (i = 0; i < nSize; i++)
	{
		nVal = fwrite(pdata, 1, WAVELEN, out);
		if (nVal > 0)
			nReturn += nVal;
		else
		{
			nReturn = -1;
			break;
		}
	}

	return nReturn;
}

//write n milliseconds of null tone
size_t writeNULL( int nDuration, FILE* out)
{
	
	const unsigned char * pdata = bSin ? ONEBIT_SIN_KCS : ONEBIT_SQR_KCS;
	size_t nVal, nReturn = 0;
	int cycles = SAMPLERATE / WAVELEN; //1225

	nDuration = (nDuration * cycles / 1000);

	while ( nDuration-- > 0)
	{
		nVal = fwrite(pdata,1,WAVELEN, out);
		if (nVal > 0)
			nReturn += nVal;
		else
		{
			nReturn = -1;
			break;
		}
	}
	return nReturn;

}

size_t WriteByte( unsigned char data, FILE *out)
{
	size_t nVal, nWrote = 0;
	char frame[12];
	int i, nFrameLen;
	unsigned char mask[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };

	frame[0] = 0; //start bit
	int parity = 0;
	for (i = 0; i < 8; i++)
	{
		if (data & mask[i])
		{
			frame[i+1] = 1;
			parity++;
		}
		else
		{
			frame[i+1] = 0;	
		}
	}
	
	if (b7bits)
	{
		nFrameLen = 8;
		//startbit, 7 data bits, parity, stop bit(s)

	}
	else
	{
		nFrameLen = 9;
		//startbit, 8 data bits, parity, stop bit(s)

	}
	//add parity bit
	if (bEven)
	{
		frame[nFrameLen++] = (parity & 1) ? 1 : 0;
	}
	else if (bOdd)
	{
		frame[nFrameLen++] = (parity & 1) ? 0 : 1;
	}
	
	frame[nFrameLen++] = 1;	  //1st stop bit;

	if (!bStop1) //2 stop bits
	{
		frame[nFrameLen++] = 1;	
	}

	for (i = 0; i < nFrameLen; i++)
	{
		nVal = writeBit( frame[i] ? 1 : 0, out);
		if (nVal <= 0)
			return -1;
		nWrote += nVal;
			
	}
	if (nWordPadding > 0)
	{
		nVal = writeNULL( nWordPadding, out);
		if (nVal > 0)
			nWrote += nVal;
	}


	return nWrote;


}

//write <CR> NUL NUL NUL ... NUL <LF>
//to allow time for BASIC parser to process line
size_t WriteBasicEOL( FILE* out)
{
	size_t nWrote = 0, nVal;
	int i;

	//Write <CR>and NULLS
	nVal = WriteByte( 0x0D, out);
	if (nVal>0)
	nWrote += nVal;
	for (i = 0; i < nNulls; i++)
	{
	 nVal = WriteByte( 0x00, out);
	 if (nVal>0)
		nWrote += nVal;

	}
	if (nPadding > 0)
	{
		nVal = writeNULL( nPadding, out);
		if (nVal > 0)
			nWrote += nVal;
	}

	 nVal = WriteByte( 0x0A, out);
	 if (nVal>0)
		nWrote += nVal;

	 return nWrote;
		
}



void usage()
{
	_fputts( _T("OSIKCS - Transcodes between Kansas City Standard audio & data file for OSI\r\n"), stdout);
	_fputts( _T("Usage: OSIKCS [options] infile [outfile]\r\n"), stdout);
	_fputts( _T("    If [outfile] is not specified, a target file is automatically\r\n")\
	         _T("       created based on input filetype\r\n"), stdout);
	_fputts( _T("    -Rn Rate for en/decoding 0=300/1=600 (default: 0)\r\n"), stdout);  
	_fputts( _T("    -F  Force overwrite of existing output file\r\n"), stdout);
	_fputts( _T("    -A  Append to output file (WAV or data)(WAV must be 22050 8bit mono)\r\n"), stdout);
	_fputts( _T("    Encode options:\r\n"), stdout);
	_fputts( _T("    -B  BASIC processing -add NULS between <CR> & <LF> like OSI BASIC\r\n"), stdout);
	_fputts( _T("    -Nn Number of Nulls (default 10)\r\n"), stdout);
	_fputts( _T("    -O  Odd parity \r\n"), stdout);
	_fputts( _T("    -E  Even parity\r\n"), stdout);
	_fputts( _T("    -D  7 Data bits\r\n"), stdout);
	_fputts( _T("    -S  1 Stop bit\r\n"), stdout);
	_fputts( _T("    -Wn padding in msec after each serial Word\r\n"), stdout);
	_fputts( _T("    -Pn Padding in msec after each EOL\r\n"), stdout);
	_fputts( _T("    -Ln Leader time (in msec) defaults to 2000\r\n"), stdout);
	_fputts( _T("    -Tn Trailer time (in msec) defaults to 1000\r\n"), stdout);
	_fputts( _T("    -U  use sine Tone instead of precompensated square wave\r\n"), stdout);
	_fputts( _T("    -Cn select WAV Channel for decode 0=Any 1=Left 2=Right\r\n"), stdout);
	_fputts( _T("    -X  Decoding output diagnostics\r\n"), stdout);
	_fputts( _T(" Default encode: create 22050 8bit mono WAV file @300 baud 8N2\r\n"), stdout);
	_fputts( _T(" Default decode: create data file from 300 baud 8N2 WAV file\r\n"), stdout);
	_fputts( _T(" (ver 1.0)\r\n"), stdout);

}

int isBigEndian() 
{
    int test = 1;
    char *p = (char*)&test;

    return p[0] == 0;
}

//append our wav output to existing wav file of the same format
int AppendWav(FILE* infile, size_t &nLen)
{

	struct _stat status = {0};
	int result = _fstat( _fileno(infile), &status );
	struct HEADER hdr;
	nLen = 0;

	if ( fread( &hdr, 1, sizeof(HEADER), infile) == sizeof(HEADER))
	{
		//some fmt_chunk_marker are "fmt " others are "fmt\0"
		if ( !strncmp(hdr.riff, "RIFF", 4) && !strncmp(hdr.wave, "WAVE", 4)  && !strncmp(hdr.fmt_chunk_marker, "fmt", 3) )
		{
			
			while (	strncmp(hdr.data_chunk_header, "data", 4))
			{
				//seek data chunk
			}
			
			
			if (!strncmp(hdr.data_chunk_header, "data", 4) )
			{
				fseek(infile, 44, SEEK_SET);

				SWAP_UINT16(hdr.channels);	 //1
				SWAP_UINT32(hdr.sample_rate);	//22050	  (blocks per second)
				SWAP_UINT32(hdr.byterate);	//22050	   bytes per sample frame times the sample rate
				SWAP_UINT16(hdr.bits_per_sample); //8
				SWAP_UINT16(hdr.block_align); //1 number of bytes required to store a single sample frame, i.e. a single sample for each channel.
				SWAP_UINT32(hdr.data_size);
				
				
				if (hdr.format_type != 1 )
				{
					return INVALID_ENCODING;
				}
				if (status.st_size < (size_t) (hdr.data_size + 44) )
				{
					return BAD_FILE;
				}
				
				printf("Existing file - Sample Rate: %d, %d bits, %d channels\r\n", hdr.sample_rate, hdr.bits_per_sample, hdr.channels);

				if (hdr.sample_rate != 22050 || hdr.bits_per_sample != 8 || hdr.channels != 1)
				{
					return INVALID_ENCODING;
				}

				//skip past existing data
				if (!fseek(infile, hdr.data_size,SEEK_CUR))
				{
					nLen = hdr.data_size;
					return 1;
				}
				return BAD_FILE;	
			}
		}
		else
		{
			return INVALID_FILETYPE;
		}

	}

	return BAD_FILE;


}

int DecodeWav(FILE* infile, FILE* out, int nChannel = 0, int rate = 0)
{

	struct _stat status = {0};
	int result = _fstat( _fileno(infile), &status );
	int nRead;
	struct HEADER hdr;

	if ( fread( &hdr, 1, sizeof(HEADER), infile) == sizeof(HEADER))
	{
		//some fmt_chunk_marker are "fmt " others are "fmt\0"
		if ( !strncmp(hdr.riff, "RIFF", 4) && !strncmp(hdr.wave, "WAVE", 4)  && !strncmp(hdr.fmt_chunk_marker, "fmt", 3) )
		{
			
			if (!strncmp(hdr.data_chunk_header, "data", 4) )
			{
				fseek(infile, 44, SEEK_SET);

				SWAP_UINT16(hdr.channels);	 //1
				SWAP_UINT32(hdr.sample_rate);	//16000	  (blocks per second)
				SWAP_UINT32(hdr.byterate);	//32000	   bytes per sample frame times the sample rate
				SWAP_UINT16(hdr.bits_per_sample); //16
				SWAP_UINT16(hdr.block_align); //2 number of bytes required to store a single sample frame, i.e. a single sample for each channel.
				SWAP_UINT32(hdr.data_size);
				
				
				if (hdr.format_type != 1 )
				{
					return INVALID_ENCODING;
				}
				if (status.st_size < (size_t) (hdr.data_size + 44) )
				{
					return BAD_FILE;
				}
			

				//sample data layout
				// 2 channel file with 16-bit pcm samples
				//      sample frame 1		         sample frame 2
				// left channel   right channel	  left channel	right channel
				// byte1 byte2	   byte1 byte2		  byte1 byte2	byte1 byte2
				long size_of_each_sample = (hdr.channels * hdr.bits_per_sample) / 8;
				long num_samples = (8 * hdr.data_size) / (hdr.channels * hdr.bits_per_sample);

				_ftprintf(logfile, _T("Decoding: %d samples/sec, %d bits %d channel(s) ~%d seconds. "), hdr.sample_rate, hdr.bits_per_sample, hdr.channels,  hdr.data_size / hdr.byterate);
				if (hdr.channels > 1 && nChannel <= hdr.channels)
				{
						_ftprintf(logfile, _T("Channel %d: "), nChannel <= 1 ? 1 : nChannel);	
				}
				


				Emitter em(out);
				eParity parity = eNone;
				if (bEven)
					parity = eEven;
				else if (bOdd)
					parity = eOdd; 

				em.setTiming(hdr.sample_rate, rate ? 600 : 300);
				em.setFrame( (b7bits ? 7 : 8), (bStop1 ? 1 : 2) , parity);


				// read each sample from data chunk if PCM
				{
					long i =0;
					unsigned char * data_buffer = new unsigned char[size_of_each_sample+2];
					int  size_is_correct = 1;

					if (!data_buffer)
					{
						return -1;
					}
					// make sure that the bytes-per-sample is completely divisible by num.of channels
					long bytes_in_each_channel = (size_of_each_sample / hdr.channels);
					if ((bytes_in_each_channel  * hdr.channels) != size_of_each_sample) 
					{
						printf("Error malformed wav: %ld x %ud <> %ld\n", bytes_in_each_channel, hdr.channels, size_of_each_sample);
						size_is_correct = 0;
					}

					if (size_is_correct) 
					{ 
								// the valid amplitude range for values based on the bits per sample
						 int low_limitL = 0;	 //0 -- 255  0 -- 65535   0 -- 	4294967295
						 int high_limitL = 0;

						 int low_limitR = 0;
						 int high_limitR = 0;

						int nLast = 0;
						bool bUp = false;
						bool bDown = false;
						bool bFirst = true;
						int tFlip = 0;
						int range = 127;
						int trigger = 6;
						int nMin = -128;

						switch (bytes_in_each_channel)
						{
							case 1:
								break;
							case 2: 
								range = 0x7FFF;
								trigger = 1638;
								low_limitR = 0;
								high_limitR = 0;
								nMin = -32768;

								break;
							case 4:
								range = 0x7FFFFFFFL;
								trigger = 6710886;
								nMin = -0x80000000L;
								break;

						}
						int iFlat = -1;
						int tFlat = -1;
						int dFlat = 0;
						int nClip = 0;

						for (i =1; i <= num_samples; i++) 
						{
							
							nRead = fread(data_buffer, size_of_each_sample, 1, infile);
							if (nRead == 1) 
							{
							
								// dump the data read
								unsigned int  xchannels = 0;
								int data = 0;

								for (xchannels = 0; xchannels < hdr.channels; xchannels ++ ) 
								{
									unsigned char * pVal = data_buffer + xchannels * bytes_in_each_channel;

									// convert data from little endian to big endian based on bytes in each channel sample
									if (bytes_in_each_channel == 4) 
									{
										SWAP_UINT32(pVal[0]);
										data = *((unsigned int *) &pVal) ;
									}
									else if (bytes_in_each_channel == 2) 
									{
										SWAP_UINT16(pVal[0]);
										data = *(( short *) pVal) ;
									}
									else if (bytes_in_each_channel == 1) 
									{
										data = pVal[0];
										data -= 128;
									}
									
									if ((xchannels == 0 && nChannel == 0) ||  (xchannels+1 == nChannel) )
									{
										

										if (iFlat > 0 && abs(data - nMin) > 3  )
										{
											//once clipping active allow for some noise
											nClip = i - iFlat-1;
											tFlat = i-1;
											iFlat = -1;
										}
										
										if ((data == nMin ) && iFlat < 0)
										{
											//watch for clipping
											iFlat = i;
											dFlat = data;
										}
										
										if (bFirst)
										{
											nLast = data;	
											bFirst = false;
										}
										else
										{
											if (data < nLast-trigger ) //down
											{
												if (!bDown)
													bDown = true;
												if (bUp)
												{
													tFlip = i;
													bUp = false;
												}
												nLast = data;

											}
											else if (data > nLast+trigger ) 
											{
												if (!bUp)
													bUp = true;
												if (bDown)
												{
													//if (bWavDiag)
													//printf(" %04d@%u ", i-tFlip, (unsigned int) i);
													if (tFlat > 0 && i-tFlat < nClip/2)
													{
														// move leading edge back to partially compensate for bottom end clipping
														int offset = nClip/3;	
														
														em.add( (size_t) i-tFlip-offset, i-offset);
														tFlat = -1;
														tFlip -= offset;
	
													}
													else
													{
														//add number of samples since last direction change
														em.add( (size_t) i-tFlip, i);
													}
													bDown = false;
												}
												nLast = data;
											}
	
										}
			
									}
									if (xchannels == 0)
									{
										if (data < low_limitL)
											low_limitL = data;
										if (data > high_limitL)
											high_limitL = data;
									}
									if (xchannels == 1)
									{
										if (data < low_limitR)
											low_limitR = data;
										if (data > high_limitR)
											high_limitR = data;

									}
								}
							}
							else 
							{
								_ftprintf( logfile, _T("Error reading file.\r\n"));
								break;
							}

						} // 	for (i =1; i <= num_samples; i++)
						if (em.GetWriteBytes() > 0)
						{
							
							_ftprintf( logfile,_T("Input frequency: %02.2fHz @ %d bps; Wrote %d bytes"),  em.GetLeaderFreq(),em.GetBPS(), em.GetWriteBytes());
							if (em.GetFEcount() > 0)
								_ftprintf( logfile,_T(" %d framing error(s)"), em.GetFEcount());
							if (em.GetPEcount() > 0)
								_ftprintf( logfile,_T(" %d parity error(s)"), em.GetPEcount());
									
							_ftprintf( logfile,_T("\r\n"));

						}
						
						if (hdr.channels == 1)
						{
							_ftprintf( logfile,_T("Sample Range: %d to %d \r\n"), low_limitL, high_limitL);
						}
						else if (hdr.channels >= 2)
						{
							_ftprintf( logfile, _T("Sample Range:L= %d to %d R= %d to %d \r\n"), low_limitL, high_limitL, low_limitR, high_limitR);
						    if (hdr.channels > 2)
								_ftprintf( logfile, _T("Additional %d channels ignored\r\n"), hdr.channels-2);
												   
						}


					} // 	if (size_is_correct) 
				
					if (data_buffer)
						delete data_buffer;
				}

				return hdr.data_size;
			}
			
		}
	}
	return -1;
}


int _tmain(int argc, TCHAR* argv[])
{
	
	FILE* inpfile = NULL;
	FILE* outfile = NULL;
	

	int nReturn = 0;
	
	TCHAR szFile[8192] = {0};



    while (--argc && ((*++argv)[0] == _T('-') )) 
	{
		const TCHAR * arg = argv[0];
		arg++;
		if (!_tcsicmp(arg, _T("o"))) 
		{          
            bOdd = true;
			continue;
        }
		if (!_tcsicmp(arg, _T("b"))) 
		{          
            bBASICmode = true;
			nNulls = 10;
			if (!nPadding)
				nPadding = 20;
			continue;
        }


		if (!_tcsicmp(arg, _T("e")))
		{
			bEven = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("d")))
		{
			b7bits = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("u")))
		{
			bSin = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("f")))
		{
			bOverwrite = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("a")))
		{
			bAppend = true;
			continue;
		}
		if (!_tcsicmp(arg, _T("s")))
		{
			bStop1 = true;
			continue;
		}
		if (!_tcsnicmp(arg, _T("r"), 1))
		{
			nRate = 0;
			if ( *(arg+1) == _T('1') )
				nRate = 1;

			continue;
		}
		if (!_tcsicmp(arg, _T("x")))
		{
			bWavDiag = true;
			continue;
		}

		
		if (!_tcsnicmp(arg, _T("p"), 1))
		{
			nPadding = _ttoi(arg+1);
			if (nPadding < 0 || nPadding > 2000)
				nPadding = 0;
			continue;
		}
		if (!_tcsnicmp(arg, _T("w"), 1))
		{
			nWordPadding = _ttoi(arg+1);
			if (nWordPadding < 0 || nWordPadding > 2000)
				nWordPadding = 0;
			continue;
		}


		if (!_tcsnicmp(arg, _T("c"), 1))
		{
			nChannel = _ttoi(arg+1);
			continue;
		}
		
		
		if (!_tcsnicmp(arg, _T("l"), 1))
		{
			nLeader = _ttoi(arg+1);
			if (nLeader < 0)
				nLeader = 2000;
			continue;
		}
		if (!_tcsnicmp(arg, _T("t"), 1))
		{
			nTrailer = _ttoi(arg+1);
			if (nTrailer < 0)
				nTrailer = 2000;
			continue;
		}
		
		if (!_tcsnicmp(arg, _T("n"), 1))
		{
			nNulls = _ttoi(arg+1);
			continue;
		}

		_fputts( _T(" Unknown argument:"), stderr);
		_fputts( arg, stderr);
		_fputts( _T("\r\n"), stderr);

		//anything else
		usage();
		return 0;
	}	
	
	if (argc > 0)
	{
	    int i = 0;

		_tcsncpy(szFile, argv[i], 8191);

		int nLen = _tcslen(szFile);
		int j = nLen-1;
		while (	j > 0 && szFile[j] != _T('.'))
			j--;
		if (j && !_tcsicmp( &szFile[j], _T(".wav")) )
		{
			bDecode = true;
		}

		if ( (inpfile = _tfopen(argv[i], _T("rb"))) == NULL)
		{
			_ftprintf(logfile, _T("osikcs: Can't open '%s' for input.\n"), argv[i]);
			return -1;
		}

		if (inpfile)
		{
			char rbuf[32];

			struct _stat buf;
			int result = _fstat( _fileno(inpfile), &buf );

			if ( buf.st_size > 32)
			{
				if (fread( &rbuf[0], sizeof( char ), 16, inpfile ) == 16)
				{
					if (!memcmp(rbuf, "RIFF", 4) )
					{
						bDecode = true;
					}
				}
				fseek(inpfile, 0, SEEK_SET);
			}
		}


		if ( ++i < argc )
		{

			if ( (outfile = _tfopen(argv[i], _T("r+b"))) != NULL )
			{
				
				if (!bOverwrite && !bAppend)
				{
					if (outfile)
					{
						fclose(outfile);
						outfile = NULL;
					}
					if (inpfile)
					{
						fclose(inpfile);
						inpfile = NULL;
					}
					_ftprintf(logfile, _T("osikcs: Target file '%s' exists. Use -f to force overwrite\n"), argv[i]);
					return -1;
				}
				if (!bAppend)
				{
					fclose(outfile);  //reopen below truncated
					outfile = NULL;
				}
			}
			if (!bAppend && (outfile = _tfopen(argv[i], _T("w+b"))) == NULL )
			{
				fclose(inpfile);
				_ftprintf(logfile, _T("osikcs: Can't open '%s' for output.\n"), argv[i]);
				return -1;
			}

		}

	}
	else
	{
		usage();
		return 0;
	}

	if (!outfile)
	{
		//make an output file base on input now that we know what file
		//we are making
		int nLen = _tcslen(szFile);
		int j = nLen-1;
		while (	j > 0 && szFile[j] != _T('.'))
			j--;
		if (szFile[j] == _T('.'))
			szFile[j] = 0;
		if (bDecode)
		{
			_tcscat(szFile, _T(".txt"));
		}
		else	
		{
			_tcscat(szFile, _T(".wav"));
		}

		if ( (outfile = _tfopen(szFile, _T("r+b"))) != NULL )
		{
			if (!bOverwrite && !bAppend)
			{
				fclose(outfile);
				fclose(inpfile);
				_ftprintf(logfile, _T("osikcs: Target file '%s' exists. Use -f to force overwrite\n"), szFile);
				goto KCSabort;
			}
			if (!bAppend)
			{
				fclose(outfile); //let it open again for truncation
				outfile = NULL;
			}

		}
		if (!outfile)
		{
			if ((outfile = _tfopen(szFile, _T("w+b"))) == NULL )
			{
				bAppend = false; 
				fclose(inpfile);
				_ftprintf(logfile, _T("osikcs: Can't open '%s' for output.\n"), szFile);
				goto KCSabort;
			}
		}

	}

	if (bDecode)
	{
		if (bAppend)
			fseek(outfile, 0, SEEK_END);
		
		int nSize = DecodeWav(inpfile, outfile, nChannel, nRate);
		if (nSize >= 0)
		{
			nReturn = 1;
		}

	}
	else
	{

		unsigned char data;
		size_t nLen =0, nWrote = 0;

		if (bAppend)
		{
			int nVal = AppendWav(outfile, nLen);
			if (nVal <= 0)
			{
				_ftprintf(logfile, _T("osikcs: Can't append to this file. "));
				if (nVal == BAD_FILE)
				{
					_ftprintf(logfile, _T("The file is corrupt.\n"));	
					
				}
				else if (nVal == INVALID_ENCODING)
				{
					_ftprintf(logfile, _T("The WAV format is not 22050/8bit mono\n"));	
				}
				else if (nVal == INVALID_FILETYPE)
				{
					_ftprintf(logfile, _T("The file is not WAV format.\n"));
				}
				else
				{
					_ftprintf(logfile, _T("\n"));
				}

				goto KCSabort;
			}
			nWrote += nLen;
		}
		else
		{
			nLen = fwrite( WAV_HDR, 1, sizeof(WAV_HDR), outfile);
			if (nLen != sizeof(WAV_HDR)) 
			{

				LogText(_T("Error writing to target file\r\n") );
				goto KCSabort;
			}
		}



		//leader
		if (nLeader > 0)
		{
			nLen = writeNULL(nLeader, outfile);
			if (nLen <= 0)
			{
				LogText(_T("Error writing to target file\r\n") );
				goto KCSabort;

			}
			nWrote += nLen;
		}
		if (bBASICmode)
		{
			//start 1st line with CR to truncate noise
			nLen = WriteBasicEOL(outfile);
			if (nLen <= 0)
			{
				LogText(_T("Error writing to target file\r\n") );
				goto KCSabort;

			}
			nWrote += nLen;

		}

		//data
		bool bCR = false;
		while (fread( &data, sizeof( char ), 1, inpfile ) == 1)
		{
			if (bCR && (data != 0x00 && data != 0x0A))
			{
				bCR = false;
			}

			if (bBASICmode) //OSI BASIC processing mode - <CR><LF> -> <CR>NUL NUL NUL... <LF>
			{
				if (bCR && data != 0x0D)  //eat NULLS and LF
				{
					continue;
				}
				if (data == 0x0D)
				{
					nLen = WriteBasicEOL(outfile);
					bCR = true;
				}
				else
				{
					nLen = WriteByte( data, outfile);
				}
			}
			else
			{

				//encode it
				nLen = WriteByte( data, outfile);
			}
			if (!bBASICmode && data == 0x0D)
			{
					if (nPadding > 0)
					{
						int nVal = writeNULL( nPadding, outfile);
						if (nVal > 0)
							nLen += nVal;
					}

			}
			if (nLen <= 0)
			{
				LogText(_T("Error writing to target file\r\n") );
				goto KCSabort;
			}
			else
				nWrote += nLen;

		}

		//trailer
		if (nTrailer > 0)
		{
			nLen = writeNULL(nTrailer, outfile);
			if (nLen > 0)
				nWrote += nLen;
		}
		nWrote++;

		fclose(inpfile); //done reading
		inpfile = NULL;

		fputc(0, outfile); //trailing NULL

		//fix chunksize in wave header
		fseek(outfile, CHUNKSIZEOFFSET, SEEK_SET);
		unsigned char chunk[4];
		
		int i;
		nLen = nWrote;
		for (i = 0; i < 4; i++)
		{
			chunk[i] = nLen & 0xFF;
			nLen = nLen >> 8;
		}
 
		fwrite(&chunk[0], 1, 4, outfile);
		
		//fix filesize chunk in wave header
		fseek(outfile, FILESIZEOFFSET, SEEK_SET);
		nWrote += CHUNKSIZEOFFSET - 4;
		for (i = 0; i < 4; i++)
		{
			chunk[i] = nWrote & 0xFF;
			nWrote = nWrote >> 8;
		}		
		fwrite(&chunk[0], 1, 4, outfile);
		nReturn = 1;
	}

KCSabort:

	if (inpfile)
		fclose(inpfile);

	if (outfile)
		fclose(outfile);
	
	return nReturn;
}
static int logpos = 0;
void LogData(unsigned char c)
{
	TCHAR hex[17] = _T("0123456789ABCDEF");
	TCHAR out[4] = {0};
    if (logfile)
	{
		
		if ((logpos%16) == 0)
		{
			_fputts( _T("\r\n"), logfile);
			logpos = 0;
		}
		logpos++;

		out[0] = hex[c >> 4];
		out[1] = hex[c &0x0F];
		out[2] = 0x20;
		out[3] = 0;
		_fputts( out, logfile);
	}
}
void LogText(const TCHAR * szFormat, ...)
{
	static TCHAR buf[2048];
    va_list vargs;

    if (logfile)
	{
		if (szFormat && *szFormat) 
		{
			va_start( vargs, szFormat );
			_vsntprintf( buf, 2047, szFormat, vargs );
			va_end( vargs );  
        
			_fputts( buf, logfile);
			logpos = 0;
		}
	}
}
